/*------------------------------------------------------------------------*/
/* Copyright (C) 2019, Daniela Kaufmann, Johannes Kepler University Linz    */
/*------------------------------------------------------------------------*/

#define VERSION "000"

static const char * usage =
"matchspec [ <option> ... ]  [ <.oldout> <.newout> <.equal> <.spec> <.polys> <.pac>]\n"
"\n"
"where <option> is one of the following\n"
"\n"
"  -h             print this command line option summary and exit\n"
"The <polynomials> argument should point to a file with the\n"
"original set of polynomials and <proof> is a path to a proof file\n"
"interpreted as a sequence of inferences in the polynomial calculus.\n"
"The tool checks that all inferences in the sequence are correct.\n"
;

/*------------------------------------------------------------------------*/
// TODO: remove powers

#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../gmp-6.1.2/gmp.h"

static FILE * pac_file;
/*------------------------------------------------------------------------*/
static void die (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr, "*** 'matchspec' error: ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  fflush (stderr);
  exit (1);
}

static void msg (const char * msg, ...) {
  fprintf (stdout, "[matchspec] ");
  va_list ap;
  va_start (ap, msg);
  vfprintf (stdout, msg, ap);
  va_end (ap);
  fprintf (stdout, "\n");
  fflush (stdout);
}

/*------------------------------------------------------------------------*/

// memory management routines maintaining a count of allocated memory

static size_t current_allocated_bytes;
static size_t max_allocated_bytes;

static void inc_allocated_bytes (size_t bytes) {
  current_allocated_bytes += bytes;
  if (current_allocated_bytes > max_allocated_bytes)
    max_allocated_bytes = current_allocated_bytes;
}

static void dec_allocated_bytes (size_t bytes) {
  assert (current_allocated_bytes >= bytes);
  current_allocated_bytes -= bytes;
}

static void * allocate (size_t bytes) {
  if (!bytes) return 0;
  void * res = malloc (bytes);
  if (bytes) {
    if (res) memset (res, 0, bytes);
    else die ("out-of-memory allocating %zu bytes", bytes);
  }
  inc_allocated_bytes (bytes);
  return res;
}

static void deallocate (void * ptr, size_t bytes) {
  dec_allocated_bytes (bytes);
  free (ptr);
}

static void * reallocate (void * ptr, size_t old_bytes, size_t new_bytes) {
  dec_allocated_bytes (old_bytes);
  void * res = realloc (ptr, new_bytes);
  if (res) {
    if (new_bytes > old_bytes)
      memset (res + old_bytes, 0, new_bytes - old_bytes);
  } else die ("out-of-memory reallocating %zu bytes", new_bytes);
  inc_allocated_bytes (new_bytes);
  return res;
}

/*------------------------------------------------------------------------*/

// 'polymorphic' array management routine maintain a count of allocated memory

#define ALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  (P) = allocate (BYTES); \
} while (0)

#define DEALLOCATE(P,N) \
do { \
  size_t BYTES = (N) * sizeof *(P); \
  deallocate ((P), BYTES); \
} while (0)

#define REALLOCATE(P,O,N) \
do { \
  size_t OLD_BYTES = (O) * sizeof *(P); \
  size_t NEW_BYTES = (N) * sizeof *(P); \
  (P) = reallocate ((P), OLD_BYTES, NEW_BYTES); \
} while (0)

/*------------------------------------------------------------------------*/

// some smaller helper macros

#define SWAP(TYPE,A,B) \
do { \
  TYPE TMP = (A); \
  (A) = (B); \
  (B) = TMP; \
} while (0)

#define MAX(A,B) \
  ((A) < (B) ? (B) : (A))

/*------------------------------------------------------------------------*/

// Random odd 64 numbers for constructing hash functions.

static uint64_t nonces[32];
static const size_t num_nonces = sizeof nonces / sizeof nonces[0];

static uint16_t rand16 () {
  int tmp = rand ();
  assert (tmp >= 0);
  uint16_t res = tmp ^ (tmp >> 16);
  return res;
}

static uint64_t rand64 () {
  uint64_t res = 0;
  for (unsigned i = 0; i < 64; i += 16)
    res |= rand16 () << i;
  return res;
}

static void init_nonces () {
  srand (42);
  for (size_t i = 0; i < num_nonces; i++)
    nonces[i] = rand64 () | 1;
}

/*------------------------------------------------------------------------*/

// String functions.

static char * duplicate_string (const char * str) {
  if (!str) return 0;
  size_t bytes = strlen (str) + 1;
  char * res = allocate (bytes);
  assert(res);
  strcpy (res, str);
  return res;
}

static void deallocate_string (char * str) {
  deallocate (str, strlen (str) + 1);
}

static uint64_t hash_string (const char * str) {
  uint64_t res = 0;
  size_t i = 0;
  for (const char * p = str; *p; p++) {
    res += *p;
    res *= nonces[i++];
    if (i == num_nonces) i = 0;
  }
  return res;
}

/*------------------------------------------------------------------------*/
typedef struct Variable Variable;

struct Variable {
  char * name;			// variable name
  uint64_t hash;		// saved hash value
  Variable * next;		// hash collision chain link
  char * pre;
  int used;
};


static uint64_t num_variables;
static uint64_t size_variables;
static Variable ** variable_table;

static void enlarge_variables () {
  uint64_t new_size_variables = size_variables ? 2*size_variables : 1;
  Variable ** new_variable_table;
  ALLOCATE (new_variable_table, new_size_variables);
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      uint64_t h = v->hash & (new_size_variables - 1);
      n = v->next;
      v->next = new_variable_table[h];
      new_variable_table[h] = v;
    }
  }
  DEALLOCATE (variable_table, size_variables);
  variable_table = new_variable_table;
  size_variables = new_size_variables;
}


static int is_valid_variable_first_letter (char ch) {
  if (ch == '_') return 1;
  if ('a' <= ch && ch <= 'z') return 1;
  if ('A' <= ch && ch <= 'Z') return 1;
  return 0;
}


static int is_valid_variable_letter (char ch) {
  if (ch == '_') return 1;
  if ('0' <= ch && ch <= '9') return 1;
  return 0;
}


static Variable * new_variable (const char * name) {
  //assert (is_valid_variable_name (name));
  if (num_variables == size_variables) enlarge_variables ();

  const uint64_t hash = hash_string (name);

  const uint64_t h = hash & (size_variables - 1);
  Variable ** p, * res;
  for (p = variable_table + h;
       (res = *p) && (strcmp (res->name, name));
       p = &res->next)
    ;
  if (!res) {
    *p = res = allocate (sizeof *res);
    res->name = duplicate_string (name);
    res->hash = hash;
    res->next = 0;
    res->used = 1;
    num_variables++;
  }
  else res->used++;
  return res;
}


static void print_variable (Variable * v, FILE * file) {
  fputs (v->name, file);
}



static int cmp_variable (Variable * a, Variable * b) {
  if (a == b) return 0;
  else return strcmp(a->name, b->name);
}



static void deallocate_variables () {
  for (uint64_t i = 0; i < size_variables; i++) {
    for (Variable * v = variable_table[i], * n; v; v = n) {
      n = v->next;
      deallocate_string (v->name);
      deallocate (v, sizeof *v);
    }
  }
  DEALLOCATE (variable_table, size_variables);
}

/*----------------------------------------------------------------------------*/
typedef struct Term Term;

struct Term {
  Variable * variable;  // variable name
  Term * rest;     // single linked list
  uint64_t ref;	   // reference counter
  uint64_t hash;   // saved hash value
  Term * next;	   // hash collision chain link
};

// Hash table to canonical store terms.
static uint64_t size_terms;
static uint64_t current_terms;
static Term ** term_table;


// compute hashvalues for terms
static uint64_t
compute_hash_term (Variable * variable, Term * rest) {
  assert (variable);
  uint64_t res = rest ? rest->hash : 0;
  res *= nonces[0];
  res += variable->hash;
  res *= nonces[1];
  return res;
}


// enlarge hash table
static void enlarge_terms () {
  uint64_t new_size_terms = size_terms ? 2*size_terms : 1;
  Term ** new_term_table;
  ALLOCATE (new_term_table, new_size_terms);
  for (uint64_t i = 0; i < size_terms; i++) {
    for (Term * m = term_table[i], * n; m; m = n) {
      uint64_t h = m->hash & (new_size_terms - 1);
      n = m->next;
      m->next = new_term_table[h];
      new_term_table[h] = m;
    }
  }
  DEALLOCATE (term_table, size_terms);
  term_table = new_term_table;
  size_terms = new_size_terms;
}


// copy a term
static Term * copy_term (Term * m) {
  if (!m) return m;
  assert(m);
  assert (m->ref > 0);

  m->ref++;
  assert (m->ref);
  return m;
}


// Then searching and adding a term extended by one power.
static Term *
new_term (Variable * variable, Term * rest) {

  if (rest) assert (cmp_variable(variable, rest->variable) > 0);
  if (current_terms == size_terms) enlarge_terms ();
  const uint64_t hash = compute_hash_term (variable, rest);
  const uint64_t h = hash & (size_terms - 1);

  Term * res;
  for (res = term_table[h];
       res && (res->variable != variable || res->rest != rest);
       res = res->next)
       ;

  if (res) {
    res->ref++; //here we extend that we found monomomial once more
  } else {
    res = allocate (sizeof *res);
    res->variable = variable;
    res->rest = copy_term (rest);
    res->ref = 1;
    res->hash = hash;
    res->next = term_table[h];
    term_table[h] = res;
    current_terms++;

  }
  return res;
}

// Decrements the reference count of a term, and actually deletes a
// term if its reference count goes to zero.  In this case it also
// removes it from the hash table and applies the same procedure to the
// suffix 'rest'.
static void deallocate_term (Term * m) {
  while (m) {
    assert (m->ref > 0);
    if (--m->ref > 0) break; //we still need m
    Term * rest = m->rest;
    const uint64_t h = m->hash & (size_terms - 1);
    Term ** p = term_table + h, * n;
    while ((n = *p) != m) assert (n), p = &n->next;
    *p = n->next;
    assert (current_terms);
    current_terms--;
    deallocate (m, sizeof *m);
    m = rest;
  }
}

// deallocate all terms
static void deallocate_terms () {
  for (uint64_t i = 0; i < size_terms; i++) {
    for (Term * m = term_table[i], *n; m; m = n) {
      n = m->next;
      assert (current_terms);
      current_terms--;
      deallocate (m, sizeof *m);
    }
  }
  DEALLOCATE (term_table, size_terms);
}

// print term
static void print_term (Term * m, FILE * file) {
  Term * tmp = m;
  while(tmp){
    print_variable(tmp->variable, file);
    tmp = tmp->rest;
    if(tmp) fputc_unlocked('*', file);
  }
}

// compare terms
static int cmp_terms(Term * t1, Term * t2){
  Term * tmp1 = t1;
  Term * tmp2 = t2;

  while(tmp1 && tmp2){
    int cmp_Variable= cmp_variable(tmp1->variable, tmp2->variable);
    if (cmp_Variable> 0) return 1;
    else if (cmp_Variable< 0) return -1;

    tmp1 = tmp1->rest;
    tmp2 = tmp2->rest;
  }
  if (tmp1) return 1;
  else if (tmp2) return -1;

  return 0;
}


/*------------------------------------------------------------------------*/
// Define a stack for sorting variables in terms
static size_t size_vstack;
static size_t num_vstack = 0;
static Variable** vstack;

// make buffer bigger
static void enlarge_vstack () {
  size_t new_size_vstack = size_vstack ? 2*size_vstack : 1;
  REALLOCATE (vstack, size_vstack, new_size_vstack);
  size_vstack = new_size_vstack;
}

// push char on buffer
static void push_vstack (Variable *v) {
  if (size_vstack == num_vstack) enlarge_vstack ();

  if (num_vstack==0) vstack[num_vstack++] = v;
  else
  {
     unsigned i = 0;
     Variable * tmp;
     while (i < num_vstack) {
       tmp = vstack[i];
       int cmp = cmp_variable(tmp, v);
       if (!cmp) return;
       if (cmp<0) break;
       i++;
     }

     for(unsigned j = num_vstack; j>i; j--)
       vstack[j] = vstack[j-1];

     vstack[i] = v;
     num_vstack++;
  }
}

// set size of buffer to zero
static void clear_vstack () { num_vstack = 0; }

// deallocate whole buffer
static void deallocate_vstack () { DEALLOCATE (vstack, size_vstack); }


// Generate a term from the variables on the stack
static Term * build_term_from_stack () {
  Term * res = 0;

  for(int j = num_vstack-1; j>=0; j--){
    Term * t = new_term(vstack[j], res);
    deallocate_term(res);
    res = t;
  }

  clear_vstack();
  return res;
}

/*------------------------------------------------------------------------*/

typedef struct Monomial Monomial;

struct Monomial {
  mpz_t coeff;
  Term * term;
};

// generate new monomial
static Monomial * new_monomial (mpz_t c, Term * m) {
 Monomial * res = allocate (sizeof *res);
 mpz_init(res->coeff);
 mpz_set(res->coeff, c);

 if (mpz_cmp_si(c, 0) == 0) res->term = 0;
 else  res->term = m;

 return res;
}


// copy a monomial
static Monomial * copy_monomial (Monomial * a) {
  assert(a);
  Monomial * t = new_monomial(a->coeff, copy_term(a->term));
  assert(t);
  return t;
}

// deallocate a monomial
static void deallocate_monomial(Monomial * t){
  mpz_clear(t->coeff);
  deallocate_term(t->term);
  deallocate(t, sizeof *t);
}

// print a monomial
static void print_monomial (Monomial * t, FILE * file, int lm){
  int sign = mpz_cmp_si(t->coeff, 0);
  if (!sign) return;
  else if (!lm && sign > 0) fputc_unlocked('+',file);

  if(!t->term)  mpz_out_str(file, 10, t->coeff);
  else if (mpz_cmp_si(t->coeff, -1) == 0) fputc_unlocked('-',file);
  else if (mpz_cmp_si(t->coeff, 1) != 0) mpz_out_str(file, 10, t->coeff);

  if (t->term) {
    if(mpz_cmp_ui(t->coeff, 1) != 0 && mpz_cmp_si(t->coeff, -1) != 0) fputc_unlocked('*',file);
    print_term(t->term, file);
  }
}

// compare two monomials
/*static int cmp_monomials(Monomial * t1, Monomial * t2){
  int term = cmp_terms(t1->term, t2->term);
  if (term) return term;

  return mpz_cmp(t1->coeff, t2->coeff);
}*/


/*------------------------------------------------------------------------*/
typedef struct Polynomial Polynomial;

struct Polynomial {
  Monomial * lm;
  Polynomial * rest; //single linked list
};

Polynomial * target;

// generate new polynomial
static Polynomial *
new_polynomial (Monomial * lm, Polynomial * rest) {
  if (rest) assert (cmp_terms(lm->term, rest->lm->term) >= 0);

  Polynomial * res = allocate (sizeof *res);
  res->lm = lm;
  res->rest = rest;

  return res;
}

static int is_one_poly (Polynomial* p){
  if(!p) return 0;
  if(p->rest) return 0;
  if(p->lm->term) return 0;
  if(mpz_cmp_si(p->lm->coeff, 1) != 0) return 0;
  return 1;
}


// deallocate a term
static void deallocate_polynomial (Polynomial * p) {
  while (p){
    deallocate_monomial(p->lm);
    Polynomial * rest = p->rest;
    deallocate (p, sizeof *p);
    p = rest;
  }
}

// print a polynomial
static void print_polynomial (Polynomial * p, FILE * file){
  if(!p) fputs("0",file);
  else if(mpz_cmp_si(p->lm->coeff, 0) == 0) fputs("0", file);
  else {
    Polynomial * tmp = p;
    int lm = 1;
    while(tmp){
      print_monomial(tmp->lm, file, lm);
      lm = 0;
      tmp=tmp->rest;
    }
  }
}

/*static int cmp_polynomials (Polynomial * tmp1, Polynomial * tmp2){

  while(tmp1 && tmp2){
    int cmp = cmp_monomials(tmp1->lm, tmp2->lm);
    if(cmp) return cmp;

    tmp1= tmp1->rest;
    tmp2= tmp2->rest;
  }
  if(tmp1) return 1;
  else if(tmp2) return -1;
  return 0;
}*/


/*------------------------------------------------------------------------*/
static size_t size_mstack;
static size_t num_mstack = 0;
static Monomial ** mstack;

// make buffer bigger
static void enlarge_mstack () {
  size_t new_size_mstack = size_mstack ? 2*size_mstack : 1;
  REALLOCATE (mstack, size_mstack, new_size_mstack);
  size_mstack = new_size_mstack;
}

// push char on buffer
static void push_mstack (Monomial * t) {
  assert(t);
  if (mpz_cmp_si(t->coeff, 0) == 0) {
    deallocate_monomial(t);
    return;
  }
  if (size_mstack == num_mstack) enlarge_mstack ();

  if (num_mstack == 0) mstack[num_mstack++] = t;

  else
  {
    assert((int)num_mstack >= 0);
    int i = num_mstack - 1;
    int cmp = -1;
    Monomial * tmp = 0;

    while (i >= 0) {
      tmp = mstack[i];
      cmp = cmp_terms(tmp->term, t->term);
      if (cmp >= 0) break;
      i--;
    }



    if (cmp == 0) {
      mpz_add(tmp->coeff, tmp->coeff, t->coeff);
      deallocate_monomial(t);

      if(mpz_cmp_si(tmp->coeff, 0) == 0){
        for(unsigned j = i; j < num_mstack-1; j++)
        mstack[j] = mstack[j+1];

        num_mstack--;
        deallocate_monomial(tmp);
      }
    }
    else {
      for (int j = num_mstack; j > i+1; j--)
        mstack[j] = mstack[j-1];

      mstack[i+1] = t;
      num_mstack++;
    }
  }
}

static void push_mstack_sorted(Monomial * t){
  assert(t);
  if (mpz_cmp_si(t->coeff, 0) == 0) {
    deallocate_monomial(t);
    return;
  }
  if (size_mstack == num_mstack) enlarge_mstack ();

  mstack[num_mstack++] = t;
}


// set size of buffer to zero
static void clear_mstack () { num_mstack = 0; }

// deallocate whole buffer
static void deallocate_mstack () { DEALLOCATE (mstack, size_mstack); }

// build a polynomial from stack
static Polynomial * build_polynomial_from_stack () {
  Polynomial * res = 0;

  for (int i = num_mstack-1; i >= 0; i--) {
    Polynomial * tmp = new_polynomial (mstack[i], res);
    res = tmp;
  }

  clear_mstack();
  return res;
}

/*------------------------------------------------------------------------*/
static unsigned size_oldpoly;
static unsigned oldpoly_count = 0;
static Polynomial ** oldpoly;

static unsigned size_newpoly;
static unsigned newpoly_count = 0;
static Polynomial ** newpoly;

static unsigned size_equalpoly;
static unsigned equalpoly_count = 0;
static Polynomial ** equalpoly;

static void enlarge_equalpoly () {
  size_t new_size_equalpoly = size_equalpoly ? 2*size_equalpoly : 1;
  REALLOCATE (equalpoly, size_equalpoly, new_size_equalpoly);
  size_equalpoly = new_size_equalpoly;
}

static void push_equalpoly (Polynomial * p) {
  if (size_equalpoly == equalpoly_count) enlarge_equalpoly ();
  equalpoly[equalpoly_count++] = p;
}

static void deallocate_equalpoly () {
  for(unsigned i = 0; i < equalpoly_count; i++){
    if(equalpoly[i]) deallocate_polynomial(equalpoly[i]);
  }
  DEALLOCATE (equalpoly, size_equalpoly);
}


static void enlarge_oldpoly () {
  size_t new_size_oldpoly = size_oldpoly ? 2*size_oldpoly : 1;
  REALLOCATE (oldpoly, size_oldpoly, new_size_oldpoly);
  size_oldpoly = new_size_oldpoly;
}

static void push_oldpoly (Polynomial * p) {
  if (size_oldpoly == oldpoly_count) enlarge_oldpoly ();
  oldpoly[oldpoly_count++] = p;
}

static void deallocate_oldpoly () {
  for(unsigned i = 0; i < oldpoly_count; i++){
    if(oldpoly[i]) deallocate_polynomial(oldpoly[i]);
  }
  DEALLOCATE (oldpoly, size_oldpoly);
}

static void enlarge_newpoly () {
  size_t new_size_newpoly = size_newpoly ? 2*size_newpoly : 1;
  REALLOCATE (newpoly, size_newpoly, new_size_newpoly);
  size_newpoly = new_size_newpoly;
}

static void push_newpoly (Polynomial * p) {
  if (size_newpoly == newpoly_count) enlarge_newpoly ();
  newpoly[newpoly_count++] = p;
}

static void deallocate_newpoly () {
  for(unsigned i = 0; i < newpoly_count; i++){
    if(newpoly[i]) deallocate_polynomial(newpoly[i]);
  }
  DEALLOCATE (newpoly, size_newpoly);
}


/*------------------------------------------------------------------------*/
// Polynomial operations: negation, addition, multiplication

// negate all monomials in a polynomial
static Polynomial * negate_poly(Polynomial * p1){
  Polynomial * res = p1;
  while(res){
    mpz_mul_si(res->lm->coeff, res->lm->coeff, -1);
    res = res->rest;
  }
  return p1;
}


//add one polynomial p2 to p1
static Polynomial * add_poly(Polynomial *p1, Polynomial *p2){
  assert(!num_mstack);

  while(p1 && p2){
    if(cmp_terms(p1->lm->term, p2->lm->term) == 1){
      push_mstack_sorted(copy_monomial(p1->lm));
      p1 = p1->rest;
    }
    else if(cmp_terms(p1->lm->term, p2->lm->term) == -1){
      push_mstack_sorted(copy_monomial(p2->lm));
      p2 = p2->rest;
    }
    else{
      Monomial * m = copy_monomial(p1->lm);
      mpz_add(m->coeff, m->coeff, p2->lm->coeff);

      if(mpz_cmp_si(m->coeff, 0) == 0){
        deallocate_monomial(m);
      } else push_mstack_sorted(m);
      p1 = p1->rest;
      p2 = p2->rest;
    }
  }

  while(p1){
    push_mstack_sorted(copy_monomial(p1->lm));
    p1 = p1->rest;
  }
  while(p2){
    push_mstack_sorted(copy_monomial(p2->lm));
    p2 = p2->rest;
  }

  Polynomial * p = build_polynomial_from_stack();
  assert(!num_mstack);
  return p;
}


// multiply two monomials
static Monomial * multiply_monomial(Monomial * t1, Monomial * t2){
  mpz_t ctmp;
  mpz_init(ctmp);
  mpz_mul(ctmp, t1->coeff, t2->coeff);

  Term * tmp1 = t1->term;
  Term * tmp2 = t2->term;

  while(tmp1 && tmp2){
    if( cmp_variable (tmp1->variable, tmp2->variable) > 0){
      push_vstack(tmp1->variable);
      tmp1 = tmp1->rest;
    } else if ( cmp_variable(tmp1->variable, tmp2->variable) < 0){
      push_vstack(tmp2->variable);
      tmp2 = tmp2->rest;
    } else {
      push_vstack(tmp1->variable);
      tmp1 = tmp1->rest;
      tmp2 = tmp2->rest;
    }
  }

  while(tmp1){
    push_vstack(tmp1->variable);
    tmp1 = tmp1->rest;
  }

  while(tmp2){
    push_vstack(tmp2->variable);
    tmp2 = tmp2->rest;
  }

  Term * m = build_term_from_stack();
  Monomial * mon = new_monomial(ctmp, copy_term(m));
  mpz_clear(ctmp);
  return mon;

}

// multiply two polynomials
static Polynomial * multiply_poly(Polynomial *p1, Polynomial *p2){
  assert(!num_mstack);

  Polynomial * tmp1, * start_over;
  if(cmp_terms(p1->lm->term, p2->lm->term)==1){
    tmp1 = p1;
    start_over = p2;
  } else {
    tmp1 = p2;
    start_over = p1;
  }

  while(tmp1){
    Polynomial * tmp2 = start_over;
    while(tmp2){
      push_mstack(multiply_monomial(tmp1->lm, tmp2->lm));
      tmp2=tmp2->rest;
    }
    tmp1=tmp1->rest;
  }

  Polynomial * p = build_polynomial_from_stack();
  return p;
}


static int contained(Term * m1, Term * m2){
  int found = 0;
  Term * tmp1 = m1;

  while(tmp1){
    Term * tmp2 = m2;
    while(tmp2){
      if(tmp1->variable == tmp2->variable) found = 1;
      tmp2 = tmp2->rest;
    }
    if(!found) return 0;
    found = 0;
    tmp1 = tmp1->rest;
  }
  return 1;
}

static int term_contained_in_Poly(Term * t, Polynomial * p){

  while(p){
    if(contained(t, p->lm->term)) return 1;
    p = p->rest;
  }
  return 0;
}




// if term m2 is contained in m1 it calculates the remainder.
static Term * term_remainder(Term * m1, Term * m2){
  assert(!num_vstack);
  Term * dividend = m1;
  Term * divisor = m2;
  assert(contained(m2, m1));

  if(m1 == m2) return 0;

  while(divisor){
    Variable* v1 = dividend->variable;
    Variable* v2 = divisor->variable;
    int cmp = cmp_variable(v1, v2);
    if(!cmp){
      divisor = divisor->rest;
      dividend = dividend->rest;
    }

    if(cmp < 0){
      push_vstack(v1);
      dividend = dividend->rest;
    }
  }

  while(dividend){
    push_vstack(dividend->variable);
    dividend = dividend->rest;
  }

  Term * m = build_term_from_stack();
  return m;
}

// we divide polynomial p1 by the lm of p2
static Polynomial * divide_lm(Polynomial * p1, Polynomial * p2){
  assert(!num_mstack);
  Term * lm = p2->lm->term;

  Polynomial * tmp = p1;

  while(tmp){
    Monomial * lm_tmp = tmp->lm;
    if(contained(lm, lm_tmp->term)){
      Term * m = term_remainder(lm_tmp->term, lm);

      mpz_t ctmp;
      mpz_init(ctmp);
      mpz_cdiv_q(ctmp, lm_tmp->coeff, p2->lm->coeff);

      push_mstack_sorted(new_monomial(ctmp, copy_term(m)));
      mpz_clear(ctmp);
    }
    tmp = tmp->rest;
  }
  Polynomial * p = build_polynomial_from_stack();
  return p;
}

static Polynomial * reduce_by_one_poly(Polynomial * p1, Polynomial * p2){
  if(!p2) return 0;
  Polynomial * negfactor = divide_lm(p1, p2);
  negfactor = negate_poly(negfactor);
  Polynomial * mult   = multiply_poly(negfactor, p2);
  Polynomial * rem    = add_poly(p1, mult);

    if(!is_one_poly(negfactor)){

      fprintf(pac_file, "d *: ");
      print_polynomial(p2, pac_file);        fprintf(pac_file, ", ");
      print_polynomial(negfactor, pac_file); fprintf(pac_file, ", ");
      print_polynomial(mult, pac_file);
      fprintf(pac_file, ";\n");
    }

    fprintf(pac_file, "d +: ");
    print_polynomial(p1, pac_file);     fprintf(pac_file, ", ");
    print_polynomial(mult, pac_file);   fprintf(pac_file, ", ");
    print_polynomial(rem, pac_file);
    fprintf(pac_file, ";\n");


  deallocate_polynomial(negfactor);
  deallocate_polynomial(mult);

  return rem;
}
/*------------------------------------------------------------------------*/
static Polynomial * reduce_by_equal_poly(Polynomial * p){
  Polynomial * eq = 0;
  for (unsigned i = 0; i < equalpoly_count; i++){
    eq = equalpoly[i];
    if(term_contained_in_Poly(eq->lm->term, p)) break;
  }
  assert(eq);
  Polynomial * res = reduce_by_one_poly(p, eq);

  return res;


}

static void resolve_pairs(){
  assert(oldpoly_count == newpoly_count);
  for (unsigned i = 0; i < oldpoly_count; i++){
    Polynomial * oldp = oldpoly[i];
    Polynomial * newp = newpoly[i];

    fprintf(pac_file, "d *: ");
    print_polynomial(newp, pac_file);        fprintf(pac_file, ", -1, ");
    Polynomial * negate = negate_poly(newp);
    print_polynomial(negate, pac_file);
    fprintf(pac_file, ";\n");

    Polynomial * add = add_poly(oldp, negate);
    fprintf(pac_file, "d +: ");
    print_polynomial(oldp, pac_file);     fprintf(pac_file, ", ");
    print_polynomial(negate, pac_file);   fprintf(pac_file, ", ");
    print_polynomial(add, pac_file);
    fprintf(pac_file, ";\n");






    if(add->rest->rest){
      Polynomial * p = reduce_by_equal_poly(add);
      deallocate_polynomial(add);
      add = p;


    }
    Polynomial * new_spec = reduce_by_one_poly(target, add);
    deallocate_polynomial(target);
    target = new_spec;

    deallocate_polynomial(add);
  }
}

/*------------------------------------------------------------------------*/

static size_t size_buffer;
static size_t num_buffer;
static char * buffer;

static void enlarge_buffer () {
  size_t new_size_buffer = size_buffer ? 2*size_buffer : 1;
  buffer = reallocate (buffer, size_buffer, new_size_buffer);
  size_buffer = new_size_buffer;
}

static void push_buffer (char ch) {
  if (size_buffer == num_buffer) enlarge_buffer ();
  buffer[num_buffer++] = ch;
}

static void clear_buffer () { num_buffer = 0; }

static void deallocate_buffer () { DEALLOCATE (buffer, size_buffer); }

/*------------------------------------------------------------------------*/

// Character reading routines.

static uint64_t lineno;
static uint64_t columno;
static uint64_t charno;
static const char * parse_file_name;
static FILE * parse_file;
static int close_parse_file;
static int saved_char;
static int char_saved;
static uint64_t last_line_length;

static int next_char () {
  int res;
  if (char_saved) {
    res = saved_char;
    char_saved = 0;
  } else {
#ifdef HAVEUNLOCKEDIO
    res = getc_unlocked (parse_file);
#else
    res = getc (parse_file);
#endif
  }
  if (res == '\n') {
    last_line_length = columno;
    columno = 0;
    lineno++;
  } else if (res != EOF) {
    columno++;
  }
  if (res != EOF) charno++;
  return res;
}

static void prev_char (int ch) {
  assert (!char_saved);
  if (ch == '\n') {
    columno = last_line_length;
    assert (lineno > 0);
    lineno--;
  } else if (ch != EOF) {
    assert (charno > 0);
    charno--;
    assert (columno > 0);
    columno--;
  }
  saved_char = ch;
  char_saved = 1;
}

/*------------------------------------------------------------------------*/

// Scanner.

typedef const char * Token;

static Token END_OF_FILE_TOKEN = "end-of-file";
static Token MINUS_TOKEN       = "minus operator";
static Token PLUS_TOKEN	       = "addition operator";
static Token MULTIPLY_TOKEN    = "multiplication operator";
static Token COLON_TOKEN       = "colon separator";
static Token COMMA_TOKEN       = "comma separator";
static Token SEMICOLON_TOKEN   = "semicolon separator";
static Token NUMBER_TOKEN      = "number";
static Token VARIABLE_TOKEN    = "variable";

static Token token;
static uint64_t lineno_at_start_of_last_token;
static uint64_t columno_at_start_of_last_token;

static void parse_error (const char * msg, ...) {
  fflush (stdout);
  fprintf (stderr,
    "*** 'matchspec' parse error in '%s'"
    " line %" PRIu64,
    parse_file_name,
    lineno_at_start_of_last_token);
  if (buffer[0] && isprint (buffer[0]))
    fprintf (stderr, " at '%s'", buffer);
  else if (token == END_OF_FILE_TOKEN)
    fputs (" at end-of-file", stderr);
  fputs (": ", stderr);
  va_list ap;
  va_start (ap, msg);
  vfprintf (stderr, msg, ap);
  va_end (ap);
  fputc ('\n', stderr);
  fflush (stderr);
  exit (1);
}

static Token new_token (Token t) {
  push_buffer (0);
  token = t;
  return token;
}

static Token next_token () {
  clear_buffer ();
  for (;;) {
    int ch = next_char ();
    if (ch == ' ' || ch == '\t' || ch == '\r' || ch == '\n')
      continue;
    lineno_at_start_of_last_token = lineno;
    columno_at_start_of_last_token = columno;
    if (ch == EOF) return new_token (END_OF_FILE_TOKEN);
    push_buffer (ch);
    if (ch == '-') return new_token (MINUS_TOKEN);
    if (ch == '+') return new_token (PLUS_TOKEN);
    if (ch == '*') return new_token (MULTIPLY_TOKEN);
    if (ch == ':') return new_token (COLON_TOKEN);
    if (ch == ',') return new_token (COMMA_TOKEN);
    if (ch == ';') return new_token (SEMICOLON_TOKEN);
    if (ch == '{') return next_token();
    if (ch == '}') return next_token();
    if ('0' <= ch && ch <= '9') {
      while ('0' <= (ch = next_char ()) && ch <= '9')
      	push_buffer (ch);
      prev_char (ch);
      return new_token (NUMBER_TOKEN);
    }
    if (is_valid_variable_first_letter (ch)) {
      while (is_valid_variable_letter (ch = next_char ()))
	       push_buffer (ch);
      prev_char (ch);
      return new_token (VARIABLE_TOKEN);
    }
    if (isprint (ch)) parse_error ("invalid character");
    else parse_error ("invalid character code 0x%02x", ch);
  }
}

static int is_separator_token (Token t) {
  if (t == COLON_TOKEN) return 1;
  if (t == COMMA_TOKEN) return 1;
  if (t == SEMICOLON_TOKEN) return 1;
  return 0;
}

/*------------------------------------------------------------------------*/

// Parser for polynomials.



static Variable * parse_variable () {
  Variable * res = new_variable (buffer);
  next_token ();
  return res;
}


static Term * parse_term () {
  while (token == VARIABLE_TOKEN) {
    Variable * variable = parse_variable ();
    push_vstack (variable);
    if (token == MULTIPLY_TOKEN) next_token ();
  }
  Term * res = build_term_from_stack();

  return res;
}

static Monomial * parse_monomial (int sign) {
  mpz_t number;
  mpz_init(number);

  if (token == NUMBER_TOKEN) {
    mpz_set_str(number, buffer, 10);
    next_token();
  }
  else if (token == VARIABLE_TOKEN) mpz_set_ui (number, 1);
  else parse_error ("expected monomial");
  if (sign < 0) {
    mpz_mul_si(number, number, -1);
  }
  if (token == MULTIPLY_TOKEN) next_token ();
  Term * term = parse_term ();

  Monomial * res = new_monomial (number, term);
  //deallocate_term (term);
  mpz_clear (number);
  return res;
}

static Polynomial * parse_polynomial () {
  int sign;
  if (token == MINUS_TOKEN) {
    next_token ();
    if (token == NUMBER_TOKEN && buffer[0] == '0')
      parse_error ("unexpected '0' after '-'");
    sign = -1;
  } else sign = 1;
  assert (!num_mstack);
  for (;;) {
    Monomial * monomial = parse_monomial (sign);


    push_mstack (monomial);
    if (is_separator_token (token)) break;
    if (token == MINUS_TOKEN) {
      sign = -1;
      next_token ();
    } else if (token == PLUS_TOKEN) {
      sign = 1;
      next_token ();
    } else parse_error ("unexpected %s", token);
  }
  Polynomial * res = build_polynomial_from_stack ();
  return res;
}

/*------------------------------------------------------------------------*/


// Original polynomial inference parsing.

static void init_parsing (const char * file_name) {
  if (file_name) {
    parse_file_name = file_name;
    parse_file = fopen (file_name, "r");
    if (!parse_file)
      die ("can not open '%s' for reading", file_name);
    close_parse_file = 1;
  } else {
    parse_file = stdin;
    parse_file_name = "<stdin>";
    close_parse_file = 0;
  }
  lineno = 1;
  columno = 0;
  charno = 0;
}

static void reset_parsing () {
  if (close_parse_file && fclose (parse_file))
    die ("failed to close '%s'", parse_file_name);
  msg ("read %" PRIu64 " bytes from '%s'", charno, parse_file_name);
}



static void parse_original_polynomial (int idx) {
  Polynomial * p = parse_polynomial ();

  if(idx == 1) push_oldpoly(p);
  else if (idx == 2) push_newpoly(p);
  else if (idx == 3) push_equalpoly(p);
  else {
    deallocate_polynomial(p);
  }

  assert (is_separator_token (token));


}

/*------------------------------------------------------------------------*/

// Parse whole files.


static void parse_spec_polynomial (const char * file_name) {
  init_parsing (file_name);
  next_token();
  msg ("reading target polynomial from '%s'", parse_file_name);
  target = parse_polynomial ();

  assert (is_separator_token (token));
  if(next_token () != END_OF_FILE_TOKEN)
    die ("unexpected %s token", token);
  reset_parsing ();
}




static void parse_original_polynomials (const char * file_name, int idx) {
  init_parsing (file_name);
  msg ("reading original polynomials from '%s'", parse_file_name);
  uint64_t original = 0;
  while (next_token () != END_OF_FILE_TOKEN)
    parse_original_polynomial (idx), original++;
  msg ("found %" PRIu64 " original polynomials in '%s'", original, parse_file_name);
  reset_parsing ();
}




/*------------------------------------------------------------------------*/

static void init () {
  init_nonces ();
}

static void deallocate_all () {
  deallocate_polynomial(target);
  deallocate_oldpoly ();
  deallocate_newpoly ();
  deallocate_equalpoly ();
  deallocate_mstack();
  deallocate_terms ();
  deallocate_vstack();
  deallocate_variables ();
  deallocate_buffer ();

  assert (!current_allocated_bytes);
}

static void reset () {
  deallocate_all ();
}

/*------------------------------------------------------------------------*/

/*------------------------------------------------------------------------*/

// Main routine for stand alone checker.

static void banner () {
  msg ("MATCHSPEC Version " VERSION);
  msg ("Copyright (C) 2019, Daniela Kaufmann, Johannes Kepler University Linz");
}

int main (int argc, char ** argv) {

  const char * oldpoly_file_name = 0;
  const char * newpoly_file_name = 0;
  const char * equal_file_name = 0;
  const char * spec_file_name = 0;
  const char * pac_file_name = 0;

  for (int i = 1; i < argc; i++) {
    if (!strcmp (argv[i], "-h") ||
        !strcmp (argv[i], "--help"))
      fputs (usage, stdout), exit (0);
    else if (argv[i][0] == '-')
      die ("invalid command line option '%s' (try '-h')", argv[i]);
    else if (pac_file_name) die ("too many command line arguments (try '-h')");
    else if (spec_file_name) pac_file_name = argv[i];
    else if (equal_file_name) spec_file_name = argv[i];
    else if (newpoly_file_name) equal_file_name = argv[i];
    else if (oldpoly_file_name) newpoly_file_name = argv[i];
    else  oldpoly_file_name = argv[i];
  }

  banner ();
  init ();

  if(!pac_file_name)die ("too few command line arguments (try '-h')");


  if (!(pac_file = fopen (pac_file_name, "a")))
      die ("can not write output to '%s'", pac_file_name);



  parse_original_polynomials (oldpoly_file_name, 1);
  parse_original_polynomials (newpoly_file_name, 2);
  parse_original_polynomials (equal_file_name, 3);
  parse_spec_polynomial (spec_file_name);

  resolve_pairs();

  reset ();
  fclose(pac_file);

  return 0;
}
